//
//	chroma noise reduction II (version 2.4) - Avisynth filter reducing noise on chroma
//	Copyright (C) 2002 Marc Fauconneau
//
//	Inspired by :
//  chroma noise reduction (version 1.1) - VirtualDub filter reducing noise on chroma
//  Copyright (C) 2000 Gilles Mouchard
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
//  Please contact me for any bugs or questions.
//  marc.fd@libertysurf.fr

//  Change log :
//         15/12/2002 - ver 2.4  - Bug fixes
//         13/11/2002 - ver 2.3  - YV12 mode, scd (scenechange detection)
//         01/08/2002 - ver 2.2  - Ugly bug fixed
//         31/07/2002 - ver 2.1  - Bug Fixes (thx to dividee ;)
//         20/07/2002 - ver 2.0  - Avisynth filter coded (from scratch)

#include <windows.h>
#include "avisynth.h"
#include <math.h>

#include <stdio.h>

#define uc unsigned char
#define ABS(X) (((X)>0)?(X):(-(X)))

class Cnr2 : public GenericVideoFilter {
	int lastn;
	uc lt[513];
	uc ut[513];
	uc vt[513];
	PVideoFrame pre;
	uc *py,*cy;
	double scdthr;
	bool log;
	void CnrInit(const char* mode, int ln, int lm, int un, int um, int vn, int vm);
public:
	Cnr2(AVSValue args, IScriptEnvironment* env);
	~Cnr2();
    PVideoFrame __stdcall GetFrame(int n, IScriptEnvironment* env);
	static AVSValue __cdecl Create(AVSValue args, void*, IScriptEnvironment* env);
};

Cnr2::Cnr2(AVSValue args, IScriptEnvironment* env) : GenericVideoFilter(args[0].AsClip())
{
	const char* mode = args[1].AsString("oxx");
	scdthr = args[2].AsFloat(8);
	log = args[9].AsBool(false);

	if (strlen(mode) != 3)
		env->ThrowError("Cnr2 : mode is a 3 chars long string");

	if (!vi.IsYUY2() && !vi.IsYV12())
		env->ThrowError("Cnr2 : Need YUY2 or YV12 colorspace");

	if (vi.IsYV12()) {
		py = new uc[vi.width*vi.height/4];
		cy = new uc[vi.width*vi.height/4];
	}

	CnrInit(mode,
			args[3].AsInt(35),args[4].AsInt(192),
			args[5].AsInt(47),args[6].AsInt(255),
			args[7].AsInt(47),args[8].AsInt(255));
}

Cnr2::~Cnr2() 
{
	if (vi.IsYV12()) {
		delete[] py;
		delete[] cy;
	}
	pre->~VideoFrame();
}


AVSValue __cdecl Cnr2::Create(AVSValue args, void* user_data, IScriptEnvironment* env)
{
	return new Cnr2(args,env);
}

void Cnr2::CnrInit(const char* mode, int ln, int lm, int un, int um, int vn, int vm) 
{
	__asm emms
	lastn = -2;
	const double pi = 3.1415926535;

	// wide/narrow - true/false
	bool Y=true,U=true,V=true;
	if (mode[0] == 'x') Y = false;
	if (mode[1] == 'x') U = false;
	if (mode[2] == 'x') V = false;

	for (int i=-256;i<256;i++) {lt[i+256] = 0;}
	for(int j = -ln; j <= ln; j++) {
		if (Y) { lt[j+256] = int( (lm/2) * ( 1 + cos((j*j*pi)/(ln*ln))));
		} else { lt[j+256] = int( (lm/2) * ( 1 + cos((j * pi)/ ln))); }
	}
	for (i=-256;i<256;i++) {ut[i+256] = 0;}
	for(j = -un; j <= un; j++) {
		if (U) { ut[j+256] = int( (um/2) * ( 1 + cos((j*j*pi)/(un*un))));
		} else { ut[j+256] = int( (um/2) * ( 1 + cos((j * pi)/ un))); }
	}
	for (i=-256;i<256;i++) {vt[i+256] = 0;}
	for(j = -vn; j <= vn; j++) {
		if (V) { vt[j+256] = int( (vm/2) * ( 1 + cos((j*j*pi)/(vn*vn))));
		} else { vt[j+256] = int( (vm/2) * ( 1 + cos((j * pi)/ vn))); }
	}
}

PVideoFrame __stdcall Cnr2::GetFrame(int n, IScriptEnvironment* env) 
{
	if (n<1) return child->GetFrame(n, env);

	PVideoFrame src = child->GetFrame(n, env);

	if (lastn+1!=n || !pre) {
		pre = child->GetFrame(n-1, env);
		env->MakeWritable(&pre);
	}

	const uc *srcp = src->GetReadPtr();
	uc *prep = pre->GetWritePtr();

	int diffacc = 0;
	int diffmax = (int)scdthr*vi.height*vi.width;

	if (vi.IsYUY2()) {
		diffmax *= 2;
		for (int fy = 0; fy < pre->GetHeight(); fy++) {
			for (int x = 0; x < pre->GetRowSize(); x+=4)
			{
				int u = srcp[x+1];
				int v = srcp[x+3];
	
				const signed int ydiff = ((srcp[x+0]+srcp[x+2]) - (prep[x+0]+prep[x+2]))>>1;
				const signed int udiff = u - (prep[x+1]);
				const signed int vdiff = v - (prep[x+3]);
				diffacc += ABS(ydiff)*2+ABS(udiff)+ABS(vdiff);
		
				const int chroma1_ratio = (lt[ydiff+256] * ut[udiff+256]);
				const int chroma2_ratio = (lt[ydiff+256] * vt[vdiff+256]);

				u = (chroma1_ratio * (prep[x+1]) + (65536 - chroma1_ratio) * u);
				v = (chroma2_ratio * (prep[x+3]) + (65536 - chroma2_ratio) * v);

				u = (u + 0x8000) >> 16;
				v = (v + 0x8000) >> 16;
	
				prep[x+0] = srcp[x+0];
				prep[x+2] = srcp[x+2];
				prep[x+1] = u;
				prep[x+3] = v;
			}
			if (diffacc>diffmax)
				goto exit;
			srcp += src->GetPitch();
			prep += pre->GetPitch();
		}
	} else {

		// YV12 is a bit slower, but and it's only a 3-4 fps difference (at 40 fps)
		
		uc* swap = py;
		py = cy;
		cy = swap;

		// downsample Y plane
		uc *cury = cy;
		int pitch = src->GetPitch();
		for (int y = 0; y < vi.height/2; y++) {
			for (int x = 0; x < vi.width/2; x++) {
				cury[x] = (srcp[x*2]+srcp[x*2+1]+srcp[x*2+pitch]+srcp[x*2+pitch+1])>>2;
			}
			srcp += pitch*2;
			cury += vi.width/2;
		}

		for (y = 0; y < pre->GetHeight(); y++)
			memcpy(	pre->GetWritePtr()+pre->GetPitch()*y,
					src->GetReadPtr()+src->GetPitch()*y,
					pre->GetRowSize());

		for (int i=2;i<=3;i++) {
			int plane = i==1?PLANAR_Y:i==2?PLANAR_U:i==3?PLANAR_V:(-1);
			uc *prey = py;
			cury = cy;
			uc *t = (plane == PLANAR_U)?ut:vt;
			srcp = src->GetReadPtr(plane);
			prep = pre->GetWritePtr(plane);
			for (int y = 0; y < pre->GetHeight(plane); y++) {
				for (int x = 0; x < pre->GetRowSize(plane); x++) 
				{
					int ydiff = cury[x]-prey[x];
					int uvdiff = srcp[x]-prep[x];
					diffacc += ABS(uvdiff)+ABS(ydiff);
					int cr = (lt[ydiff+256] * t[uvdiff+256]);
					prep[x] = ((cr * (prep[x]) + (65536 - cr) * srcp[x]) + 0x8000) >> 16;
				}
				if (diffacc>diffmax)
					goto exit;

				srcp += src->GetPitch(plane);
				prep += pre->GetPitch(plane);
				cury += vi.width/2;
				prey += vi.width/2;
			}
		}
	}

exit:

	if (log) {
		char buffer[255];
		sprintf(buffer,"Cnr2 log> acc/max %d/%d  \t ratio %f",diffacc,diffmax,diffacc/(double)diffmax);
		OutputDebugString(buffer);
	}

	if (diffacc>diffmax)
		pre = src;

	lastn = n;
	return pre;
}

extern "C" __declspec(dllexport) const char* __stdcall AvisynthPluginInit(IScriptEnvironment* env) {
    env->AddFunction("Cnr2", "c[mode]s[scdthr]f[ln]i[lm]i[un]i[um]i[vn]i[vm]i[log]b", Cnr2::Create, 0);
    return "`Cnr2' Chroma Noise Reducer 2.4";
}